@@ -19,19 +19,17 @@
  */
 
 #include <cstdio>
-#include <cstdlib>
 
 #include <string>
 
 #include "common.h"
-#include "s3fs.h"
 #include "s3fs_help.h"
 #include "s3fs_auth.h"
 
 //-------------------------------------------------------------------
 // Contents
 //-------------------------------------------------------------------
-static const char help_string[] = 
+static constexpr char help_string[] =
     "\n"
     "Mount an Amazon S3 bucket as a file system.\n"
     "\n"
@@ -87,7 +85,7 @@ static const char help_string[] =
     "   storage_class (default=\"standard\")\n"
     "      - store object with specified storage class. Possible values:\n"
     "        standard, standard_ia, onezone_ia, reduced_redundancy,\n"
-    "        intelligent_tiering, glacier, and deep_archive.\n"
+    "        intelligent_tiering, glacier, glacier_ir, and deep_archive.\n"
     "\n"
     "   use_rrs (default is disable)\n"
     "      - use Amazon's Reduced Redundancy Storage.\n"
@@ -130,6 +128,8 @@ static const char help_string[] =
     "        environment which value is <kms id>. You must be careful\n"
     "        about that you can not use the KMS id which is not same EC2\n"
     "        region.\n"
+    "        Additionally, if you specify SSE-KMS, your endpoints must use\n"
+    "        Secure Sockets Layer(SSL) or Transport Layer Security(TLS).\n"
     "\n"
     "   load_sse_c - specify SSE-C keys\n"
     "        Specify the custom-provided encryption keys file path for decrypting\n"
@@ -194,25 +194,28 @@ static const char help_string[] =
     "\n"
     "   stat_cache_expire (default is 900))\n"
     "      - specify expire time (seconds) for entries in the stat cache.\n"
-    "        This expire time indicates the time since stat cached. and this\n"
-    "        is also set to the expire time of the symbolic link cache.\n"
+    "        This expire time indicates the time since stat cached.\n"
     "\n"
     "   stat_cache_interval_expire (default is 900)\n"
-    "      - specify expire time (seconds) for entries in the stat cache(and\n"
-    "        symbolic link cache).\n"
+    "      - specify expire time (seconds) for entries in the stat cache.\n"
     "      This expire time is based on the time from the last access time\n"
     "      of the stat cache. This option is exclusive with stat_cache_expire,\n"
     "      and is left for compatibility with older versions.\n"
     "\n"
-    "   enable_noobj_cache (default is disable)\n"
-    "      - enable cache entries for the object which does not exist.\n"
-    "      s3fs always has to check whether file (or sub directory) exists \n"
-    "      under object (path) when s3fs does some command, since s3fs has \n"
-    "      recognized a directory which does not exist and has files or \n"
-    "      sub directories under itself. It increases ListBucket request \n"
-    "      and makes performance bad.\n"
-    "      You can specify this option for performance, s3fs memorizes \n"
-    "      in stat cache that the object (file or directory) does not exist.\n"
+    "   enable_negative_cache (default is enabled negative cache)\n"
+    "      - This option will keep non-existence of objects in a stat cache.\n"
+    "        When this negative cache is enabled, it will not process extra\n"
+    "        HeadObject requests to search for non-existent objects, improving\n"
+    "        performance.\n"
+    "        This feature is enabled by default, so there is no need to specify\n"
+    "        it.\n"
+    "\n"
+    "   disable_negative_cache (default is enabled negative cache)\n"
+    "      - By default, s3fs keeps non-existent objects in the stat cache.\n"
+    "        This option disables this negative caching.\n"
+    "        This prevents delays in updates due to cache retention.\n"
+    "        However, it may increase the number of HeadObject requests to check\n"
+    "        if an object exists, which may decrease performance.\n"
     "\n"
     "   no_check_certificate\n"
     "      - server certificate won't be checked against the available \n"
@@ -221,6 +224,21 @@ static const char help_string[] =
     "   ssl_verify_hostname (default=\"2\")\n"
     "      - When 0, do not verify the SSL certificate against the hostname.\n"
     "\n"
+    "   ssl_client_cert (default=\"\")\n"
+    "      - Specify an SSL client certificate.\n"
+    "      Specify this optional parameter in the following format:\n"
+    "      \"<SSL Cert>[:<Cert Type>[:<Private Key>[:<Key Type>\n"
+    "                                                   [:<Password>]]]]\"\n"
+    "      <SSL Cert>:    Client certificate.\n"
+    "                     Specify the file path or NickName(for NSS, etc.).\n"
+    "      <Cert Type>:   Type of certificate, default is \"PEM\"(optional).\n"
+    "      <Private Key>: Certificate's private key file(optional).\n"
+    "      <Key Type>:    Type of private key, default is \"PEM\"(optional).\n"
+    "      <Password>:    Passphrase of the private key(optional).\n"
+    "                     It is also possible to omit this value and specify\n"
+    "                     it using the environment variable\n"
+    "                     \"S3FS_SSL_PRIVKEY_PASSWORD\".\n"
+    "\n"
     "   nodnscache (disable DNS cache)\n"
     "      - s3fs is always using DNS cache, this option make DNS cache disable.\n"
     "\n"
@@ -228,17 +246,6 @@ static const char help_string[] =
     "      - s3fs is always using SSL session cache, this option make SSL \n"
     "      session cache disable.\n"
     "\n"
-    "   multireq_max (default=\"20\")\n"
-    "      - maximum number of parallel request for listing objects.\n"
-    "\n"
-    "   parallel_count (default=\"5\")\n"
-    "      - number of parallel request for uploading big objects.\n"
-    "      s3fs uploads large object (over 20MB) by multipart post request, \n"
-    "      and sends parallel requests.\n"
-    "      This option limits parallel request count which s3fs requests \n"
-    "      at once. It is necessary to set this value depending on a CPU \n"
-    "      and a network band.\n"
-    "\n"
     "   multipart_size (default=\"10\")\n"
     "      - part size, in MB, for each multipart request.\n"
     "      The minimum value is 5 MB and the maximum value is 5 GB.\n"
@@ -255,6 +262,19 @@ static const char help_string[] =
     "      The minimum value is 50 MB. -1 value means disable.\n"
     "      Cannot be used with nomixupload.\n"
     "\n"
+    "   bucket_size (default=maximum long unsigned integer value)\n"
+    "      - The size of the bucket with which the corresponding\n"
+    "      elements of the statvfs structure will be filled. The option\n"
+    "      argument is an integer optionally followed by a\n"
+    "      multiplicative suffix (GB, GiB, TB, TiB, PB, PiB,\n"
+    "      EB, EiB) (no spaces in between). If no suffix is supplied,\n"
+    "      bytes are assumed; eg: 20000000, 30GB, 45TiB. Note that\n"
+    "      s3fs does not compute the actual volume size (too\n"
+    "      expensive): by default it will assume the maximum possible\n"
+    "      size; however, since this may confuse other software which\n"
+    "      uses s3fs, the advertised bucket size can be set with this\n"
+    "      option.\n"
+    "\n"
     "   ensure_diskfree (default 0)\n"
     "      - sets MB to ensure disk free space. This option means the\n"
     "        threshold of free space size on disk which is used for the\n"
@@ -263,6 +283,15 @@ static const char help_string[] =
     "        space is smaller than this value, s3fs do not use disk space\n"
     "        as possible in exchange for the performance.\n"
     "\n"
+    "   free_space_ratio (default=\"10\")\n"
+    "      - sets min free space ratio of the disk.\n"
+    "      The value of this option can be between 0 and 100. It will control\n"
+    "      the size of the cache according to this ratio to ensure that the\n"
+    "      idle ratio of the disk is greater than this value.\n"
+    "      For example, when the disk space is 50GB, the default value will\n"
+    "      ensure that the disk will reserve at least 50GB * 10%% = 5GB of\n"
+    "      remaining space.\n"
+    "\n"
     "   multipart_threshold (default=\"25\")\n"
     "      - threshold, in MB, to use multipart upload instead of\n"
     "        single-part. Must be at least 5 MB.\n"
@@ -283,8 +312,8 @@ static const char help_string[] =
     "        If you do not use https, please specify the URL with the url\n"
     "        option.\n"
     "\n"
-    "   endpoint (default=\"us-east-1\")\n"
-    "      - sets the endpoint to use on signature version 4\n"
+    "   region (default=\"us-east-1\")\n"
+    "      - sets the region to use on signature version 4\n"
     "      If this option is not specified, s3fs uses \"us-east-1\" region as\n"
     "      the default. If the s3fs could not connect to the region specified\n"
     "      by this option, s3fs could not run. But if you do not specify this\n"
@@ -292,6 +321,7 @@ static const char help_string[] =
     "      will retry to automatically connect to the other region. So s3fs\n"
     "      can know the correct region name, because s3fs can find it in an\n"
     "      error from the S3 server.\n"
+    "      You can also specify the legacy -o endpoint which means the same thing.\n"
     "\n"
     "   sigv2 (default is signature version 4 falling back to version 2)\n"
     "      - sets signing AWS requests by using only signature version 2\n"
@@ -314,6 +344,23 @@ static const char help_string[] =
     "\n"
     "   nomultipart (disable multipart uploads)\n"
     "\n"
+    "   streamupload (default is disable)\n"
+    "      - Enable stream upload.\n"
+    "      If this option is enabled, a sequential upload will be performed\n"
+    "      in parallel with the write from the part that has been written\n"
+    "      during a multipart upload.\n"
+    "      This is expected to give better performance than other upload\n"
+    "      functions.\n"
+    "      Note that this option is still experimental and may change in the\n"
+    "      future.\n"
+    "\n"
+    "   max_thread_count (default is \"10\")\n"
+    "      - This value is the maximum number of parallel requests to be\n"
+    "      sent, and the number of parallel processes for head requests,\n"
+    "      multipart uploads and stream uploads.\n"
+    "      Worker threads will be started to process requests according to\n"
+    "      this value.\n"
+    "\n"
     "   enable_content_md5 (default is disable)\n"
     "      - Allow S3 server to check data integrity of uploads via the\n"
     "      Content-MD5 header. This can add CPU overhead to transfers.\n"
@@ -347,6 +394,22 @@ static const char help_string[] =
     "   ibm_iam_endpoint (default is https://iam.cloud.ibm.com)\n"
     "      - sets the URL to use for IBM IAM authentication.\n"
     "\n"
+    "   credlib (default=\"\" which means disabled)\n"
+    "      - Specifies the shared library that handles the credentials\n"
+    "      containing the authentication token.\n"
+    "      If this option is specified, the specified credential and token\n"
+    "      processing provided by the shared library ant will be performed\n"
+    "      instead of the built-in credential processing.\n"
+    "      This option cannot be specified with passwd_file, profile,\n"
+    "      use_session_token, ecs, ibm_iam_auth, ibm_iam_endpoint, imdsv1only\n"
+    "      and iam_role option.\n"
+    "\n"
+    "   credlib_opts (default=\"\" which means disabled)\n"
+    "      - Specifies the options to pass when the shared library specified\n"
+    "      in credlib is loaded and then initialized.\n"
+    "      For the string specified in this option, specify the string defined\n"
+    "      by the shared library.\n"
+    "\n"
     "   use_xattr (default is not handling the extended attribute)\n"
     "      Enable to handle the extended attribute (xattrs).\n"
     "      If you set this option, you can use the extended attribute.\n"
@@ -416,23 +479,20 @@ static const char help_string[] =
     "        for a object, then the object will not be able to be allowed to\n"
     "        list/modify.\n"
     "\n"
-    "   notsup_compat_dir (disable support of alternative directory names)\n"
-    "        s3fs supports the three different naming schemas \"dir/\",\n"
-    "        \"dir\" and \"dir_$folder$\" to map directory names to S3\n"
-    "        objects and vice versa. As a fourth variant, directories can be\n"
+    "   compat_dir (enable support of alternative directory names)\n"
+    "        s3fs supports two different naming schemas \"dir/\" and\n"
+    "        \"dir\" to map directory names to S3 objects and\n"
+    "        vice versa by default. As a third variant, directories can be\n"
     "        determined indirectly if there is a file object with a path (e.g.\n"
     "        \"/dir/file\") but without the parent directory.\n"
+    "        This option enables a fourth variant, \"dir_$folder$\", created by\n"
+    "        older applications.\n"
     "        \n"
     "        S3fs uses only the first schema \"dir/\" to create S3 objects for\n"
     "        directories."
     "        \n"
     "        The support for these different naming schemas causes an increased\n"
     "        communication effort.\n"
-    "        \n"
-    "        If all applications exclusively use the \"dir/\" naming scheme and\n"
-    "        the bucket does not contain any objects with a different naming \n"
-    "        scheme, this option can be used to disable support for alternative\n"
-    "        naming schemes. This reduces access time and can save costs.\nq"
     "\n"
     "   use_wtf8 - support arbitrary file system encoding.\n"
     "        S3 requires all object names to be valid UTF-8. But some\n"
@@ -460,6 +520,34 @@ static const char help_string[] =
     "        If this file does not exist on macOS, then \"/etc/apache2/mime.types\"\n"
     "        is checked as well.\n"
     "\n"
+    "   proxy (default=\"\")\n"
+    "        This option specifies a proxy to S3 server.\n"
+    "        Specify the proxy with '[<scheme://]hostname(fqdn)[:<port>]' formatted.\n"
+    "        '<schema>://' can be omitted, and 'http://' is used when omitted.\n"
+    "        Also, ':<port>' can also be omitted. If omitted, port 443 is used for\n"
+    "        HTTPS schema, and port 1080 is used otherwise.\n"
+    "        This option is the same as the curl command's '--proxy(-x)' option and\n"
+    "        libcurl's 'CURLOPT_PROXY' flag.\n"
+    "        This option is equivalent to and takes precedence over the environment\n"
+    "        variables 'http_proxy', 'all_proxy', etc.\n"
+    "\n"
+    "   proxy_cred_file (default=\"\")\n"
+    "        This option specifies the file that describes the username and\n"
+    "        passphrase for authentication of the proxy when the HTTP schema\n"
+    "        proxy is specified by the 'proxy' option.\n"
+    "        Username and passphrase are valid only for HTTP schema. If the HTTP\n"
+    "        proxy does not require authentication, this option is not required.\n"
+    "        Separate the username and passphrase with a ':' character and\n"
+    "        specify each as a URL-encoded string.\n"
+    "\n"
+    "   ipresolve (default=\"whatever\")\n"
+    "        Select what type of IP addresses to use when establishing a\n"
+    "        connection.\n"
+    "        Default('whatever') can use addresses of all IP versions(IPv4 and\n"
+    "        IPv6) that your system allows. If you specify 'IPv4', only IPv4\n"
+    "        addresses are used. And when 'IPv6'is specified, only IPv6 addresses\n"
+    "        will be used.\n"
+    "\n"
     "   logfile - specify the log output file.\n"
     "        s3fs outputs the log file to syslog. Alternatively, if s3fs is\n"
     "        started with the \"-f\" option specified, the log will be output\n"
@@ -499,6 +587,12 @@ static const char help_string[] =
     "        check result to that file. The file path parameter can be omitted.\n"
     "        If omitted, the result will be output to stdout or syslog.\n"
     "\n"
+    "   update_parent_dir_stat (default is disable)\n"
+    "        The parent directory's mtime and ctime are updated when a file or\n"
+    "        directory is created or deleted (when the parent directory's inode is\n"
+    "        updated).\n"
+    "        By default, parent directory statistics are not updated.\n"
+    "\n"
     "FUSE/mount Options:\n"
     "\n"
     "   Most of the generic mount options described in 'man mount' are\n"
@@ -557,7 +651,7 @@ void show_help()
 void show_version()
 {
     printf(
-    "Amazon Simple Storage Service File System V%s (commit:%s) with %s\n"
+    "Amazon Simple Storage Service File System V%s%s with %s\n"
     "Copyright (C) 2010 Randy Rizun <rrizun@gmail.com>\n"
     "License GPL2: GNU GPL version 2 <https://gnu.org/licenses/gpl.html>\n"
     "This is free software: you are free to change and redistribute it.\n"
@@ -567,7 +661,7 @@ void show_version()
 
 const char* short_version()
 {
-    static const char short_ver[] = "s3fs version " VERSION "(" COMMIT_HASH_VAL ")";
+    static constexpr char short_ver[] = "s3fs version " VERSION "" COMMIT_HASH_VAL;
     return short_ver;
 }
 
